@@ -16,19 +16,27 @@
  *
  * The above copyright notice and this permission notice shall be included in
  * all copies or substantial portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include <babeltrace/ctf-text/types.h>
 #include <stdio.h>
 
-int ctf_text_sequence_write(struct stream_pos *ppos, struct definition *definition)
+int ctf_text_sequence_write(struct bt_stream_pos *ppos, struct bt_definition *definition)
 {
 	struct ctf_text_stream_pos *pos = ctf_text_pos(ppos);
 	struct definition_sequence *sequence_definition =
 		container_of(definition, struct definition_sequence, p);
 	struct declaration_sequence *sequence_declaration =
 		sequence_definition->declaration;
-	struct declaration *elem = sequence_declaration->elem;
+	struct bt_declaration *elem = sequence_declaration->elem;
 	int field_nr_saved;
 	int ret = 0;
 
@@ -41,7 +49,7 @@ int ctf_text_sequence_write(struct stream_pos *ppos, struct definition *definiti
 		fprintf(pos->fp, " ");
 		if (pos->print_names)
 			fprintf(pos->fp, "%s = ",
-				g_quark_to_string(definition->name));
+				rem_(g_quark_to_string(definition->name)));
 	}
 
 	if (elem->id == CTF_TYPE_INTEGER) {
@@ -55,7 +63,7 @@ int ctf_text_sequence_write(struct stream_pos *ppos, struct definition *definiti
 			    && integer_declaration->p.alignment == CHAR_BIT)) {
 				pos->string = sequence_definition->string;
 				g_string_assign(sequence_definition->string, "");
-				ret = sequence_rw(ppos, definition);
+				ret = bt_sequence_rw(ppos, definition);
 				pos->string = NULL;
 			}
 			fprintf(pos->fp, "\"%s\"", sequence_definition->string->str);
@@ -69,7 +77,7 @@ int ctf_text_sequence_write(struct stream_pos *ppos, struct definition *definiti
 	}
 	field_nr_saved = pos->field_nr;
 	pos->field_nr = 0;
-	ret = sequence_rw(ppos, definition);
+	ret = bt_sequence_rw(ppos, definition);
 	if (!pos->dummy) {
 		pos->depth--;
 		fprintf(pos->fp, " ]");
