@@ -14,6 +14,14 @@
  *
  * The above copyright notice and this permission notice shall be included in
  * all copies or substantial portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include <stdio.h>
@@ -30,8 +38,8 @@
 #include "ctf-parser.h"
 #include "ctf-ast.h"
 
-#define _cds_list_first_entry(ptr, type, member)	\
-	cds_list_entry((ptr)->next, type, member)
+#define _bt_list_first_entry(ptr, type, member)	\
+	bt_list_entry((ptr)->next, type, member)
 
 #define fprintf_dbg(fd, fmt, args...)	fprintf(fd, "%s: " fmt, __func__, ## args)
 
@@ -47,7 +55,7 @@ int ctf_visitor_unary_expression(FILE *fd, int depth, struct ctf_node *node)
 	switch (node->parent->type) {
 	case NODE_CTF_EXPRESSION:
 		is_ctf_exp = 1;
-		cds_list_for_each_entry(iter, &node->parent->u.ctf_expression.left,
+		bt_list_for_each_entry(iter, &node->parent->u.ctf_expression.left,
 					siblings) {
 			if (iter == node) {
 				is_ctf_exp_left = 1;
@@ -107,7 +115,10 @@ int ctf_visitor_unary_expression(FILE *fd, int depth, struct ctf_node *node)
 	case NODE_ROOT:
 	case NODE_EVENT:
 	case NODE_STREAM:
+	case NODE_ENV:
 	case NODE_TRACE:
+	case NODE_CLOCK:
+	case NODE_CALLSITE:
 	case NODE_TYPEDEF:
 	case NODE_TYPEALIAS_TARGET:
 	case NODE_TYPEALIAS_ALIAS:
@@ -127,7 +138,7 @@ int ctf_visitor_unary_expression(FILE *fd, int depth, struct ctf_node *node)
 	switch (node->u.unary_expression.link) {
 	case UNARY_LINK_UNKNOWN:
 		/* We don't allow empty link except on the first node of the list */
-		if (is_ctf_exp && _cds_list_first_entry(is_ctf_exp_left ?
+		if (is_ctf_exp && _bt_list_first_entry(is_ctf_exp_left ?
 					  &node->parent->u.ctf_expression.left :
 					  &node->parent->u.ctf_expression.right,
 					  struct ctf_node,
@@ -152,7 +163,7 @@ int ctf_visitor_unary_expression(FILE *fd, int depth, struct ctf_node *node)
 			goto errperm;
 		}
 		/* We don't allow link on the first node of the list */
-		if (is_ctf_exp && _cds_list_first_entry(is_ctf_exp_left ?
+		if (is_ctf_exp && _bt_list_first_entry(is_ctf_exp_left ?
 					  &node->parent->u.ctf_expression.left :
 					  &node->parent->u.ctf_expression.right,
 					  struct ctf_node,
@@ -168,7 +179,7 @@ int ctf_visitor_unary_expression(FILE *fd, int depth, struct ctf_node *node)
 			goto errperm;
 		}
 		/* We don't allow link on the first node of the list */
-		if (_cds_list_first_entry(&node->parent->u.enumerator.values,
+		if (_bt_list_first_entry(&node->parent->u.enumerator.values,
 					  struct ctf_node,
 					  siblings) == node) {
 			fprintf(fd, "[error]: semantic error (link \"...\" is not allowed on the first node of the unary expression list)\n");
@@ -209,7 +220,10 @@ int ctf_visitor_type_specifier_list(FILE *fd, int depth, struct ctf_node *node)
 
 	case NODE_EVENT:
 	case NODE_STREAM:
+	case NODE_ENV:
 	case NODE_TRACE:
+	case NODE_CLOCK:
+	case NODE_CALLSITE:
 	case NODE_UNARY_EXPRESSION:
 	case NODE_TYPEALIAS:
 	case NODE_TYPE_SPECIFIER:
@@ -248,7 +262,10 @@ int ctf_visitor_type_specifier(FILE *fd, int depth, struct ctf_node *node)
 	case NODE_ROOT:
 	case NODE_EVENT:
 	case NODE_STREAM:
+	case NODE_ENV:
 	case NODE_TRACE:
+	case NODE_CLOCK:
+	case NODE_CALLSITE:
 	case NODE_UNARY_EXPRESSION:
 	case NODE_TYPEALIAS:
 	case NODE_TYPE_SPECIFIER:
@@ -282,7 +299,7 @@ int ctf_visitor_type_declarator(FILE *fd, int depth, struct ctf_node *node)
 		/*
 		 * A nested type declarator is not allowed to contain pointers.
 		 */
-		if (!cds_list_empty(&node->u.type_declarator.pointers))
+		if (!bt_list_empty(&node->u.type_declarator.pointers))
 			goto errperm;
 		break;			/* OK */
 	case NODE_TYPEALIAS_TARGET:
@@ -301,7 +318,7 @@ int ctf_visitor_type_declarator(FILE *fd, int depth, struct ctf_node *node)
 		 */
 		if (node->u.type_declarator.type == TYPEDEC_NESTED)
 			goto errperm;
-		cds_list_for_each_entry(iter, &node->parent->u.typealias_alias.type_specifier_list->u.type_specifier_list.head,
+		bt_list_for_each_entry(iter, &node->parent->u.typealias_alias.type_specifier_list->u.type_specifier_list.head,
 					siblings) {
 			switch (iter->u.type_specifier.type) {
 			case TYPESPEC_FLOATING_POINT:
@@ -310,7 +327,7 @@ int ctf_visitor_type_declarator(FILE *fd, int depth, struct ctf_node *node)
 			case TYPESPEC_STRUCT:
 			case TYPESPEC_VARIANT:
 			case TYPESPEC_ENUM:
-				if (cds_list_empty(&node->u.type_declarator.pointers))
+				if (bt_list_empty(&node->u.type_declarator.pointers))
 					goto errperm;
 				break;
 			default:
@@ -328,7 +345,10 @@ int ctf_visitor_type_declarator(FILE *fd, int depth, struct ctf_node *node)
 	case NODE_ROOT:
 	case NODE_EVENT:
 	case NODE_STREAM:
+	case NODE_ENV:
 	case NODE_TRACE:
+	case NODE_CLOCK:
+	case NODE_CALLSITE:
 	case NODE_CTF_EXPRESSION:
 	case NODE_UNARY_EXPRESSION:
 	case NODE_TYPEALIAS:
@@ -345,7 +365,7 @@ int ctf_visitor_type_declarator(FILE *fd, int depth, struct ctf_node *node)
 		goto errinval;
 	}
 
-	cds_list_for_each_entry(iter, &node->u.type_declarator.pointers,
+	bt_list_for_each_entry(iter, &node->u.type_declarator.pointers,
 				siblings) {
 		ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 		if (ret)
@@ -364,7 +384,7 @@ int ctf_visitor_type_declarator(FILE *fd, int depth, struct ctf_node *node)
 				return ret;
 		}
 		if (!node->u.type_declarator.u.nested.abstract_array) {
-			cds_list_for_each_entry(iter, &node->u.type_declarator.u.nested.length,
+			bt_list_for_each_entry(iter, &node->u.type_declarator.u.nested.length,
 						siblings) {
 				if (iter->type != NODE_UNARY_EXPRESSION) {
 					fprintf(fd, "[error] %s: expecting unary expression as length\n", __func__);
@@ -414,24 +434,27 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 	int ret = 0;
 	struct ctf_node *iter;
 
+	if (node->visited)
+		return 0;
+
 	switch (node->type) {
 	case NODE_ROOT:
-		cds_list_for_each_entry(iter, &node->u.root.declaration_list, siblings) {
+		bt_list_for_each_entry(iter, &node->u.root.declaration_list, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
 		}
-		cds_list_for_each_entry(iter, &node->u.root.trace, siblings) {
+		bt_list_for_each_entry(iter, &node->u.root.trace, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
 		}
-		cds_list_for_each_entry(iter, &node->u.root.stream, siblings) {
+		bt_list_for_each_entry(iter, &node->u.root.stream, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
 		}
-		cds_list_for_each_entry(iter, &node->u.root.event, siblings) {
+		bt_list_for_each_entry(iter, &node->u.root.event, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -446,7 +469,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 			goto errinval;
 		}
 
-		cds_list_for_each_entry(iter, &node->u.event.declaration_list, siblings) {
+		bt_list_for_each_entry(iter, &node->u.event.declaration_list, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -460,7 +483,21 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 			goto errinval;
 		}
 
-		cds_list_for_each_entry(iter, &node->u.stream.declaration_list, siblings) {
+		bt_list_for_each_entry(iter, &node->u.stream.declaration_list, siblings) {
+			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
+			if (ret)
+				return ret;
+		}
+		break;
+	case NODE_ENV:
+		switch (node->parent->type) {
+		case NODE_ROOT:
+			break;			/* OK */
+		default:
+			goto errinval;
+		}
+
+		bt_list_for_each_entry(iter, &node->u.env.declaration_list, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -474,7 +511,35 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 			goto errinval;
 		}
 
-		cds_list_for_each_entry(iter, &node->u.trace.declaration_list, siblings) {
+		bt_list_for_each_entry(iter, &node->u.trace.declaration_list, siblings) {
+			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
+			if (ret)
+				return ret;
+		}
+		break;
+	case NODE_CLOCK:
+		switch (node->parent->type) {
+		case NODE_ROOT:
+			break;			/* OK */
+		default:
+			goto errinval;
+		}
+
+		bt_list_for_each_entry(iter, &node->u.clock.declaration_list, siblings) {
+			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
+			if (ret)
+				return ret;
+		}
+		break;
+	case NODE_CALLSITE:
+		switch (node->parent->type) {
+		case NODE_ROOT:
+			break;			/* OK */
+		default:
+			goto errinval;
+		}
+
+		bt_list_for_each_entry(iter, &node->u.callsite.declaration_list, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -486,7 +551,10 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		case NODE_ROOT:
 		case NODE_EVENT:
 		case NODE_STREAM:
+		case NODE_ENV:
 		case NODE_TRACE:
+		case NODE_CLOCK:
+		case NODE_CALLSITE:
 		case NODE_FLOATING_POINT:
 		case NODE_INTEGER:
 		case NODE_STRING:
@@ -512,12 +580,12 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		}
 
 		depth++;
-		cds_list_for_each_entry(iter, &node->u.ctf_expression.left, siblings) {
+		bt_list_for_each_entry(iter, &node->u.ctf_expression.left, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
 		}
-		cds_list_for_each_entry(iter, &node->u.ctf_expression.right, siblings) {
+		bt_list_for_each_entry(iter, &node->u.ctf_expression.right, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -553,6 +621,9 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		case NODE_STRING:
 		case NODE_ENUMERATOR:
 		case NODE_ENUM:
+		case NODE_CLOCK:
+		case NODE_CALLSITE:
+		case NODE_ENV:
 		default:
 			goto errinval;
 		}
@@ -562,7 +633,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 			node->u._typedef.type_specifier_list);
 		if (ret)
 			return ret;
-		cds_list_for_each_entry(iter, &node->u._typedef.type_declarators, siblings) {
+		bt_list_for_each_entry(iter, &node->u._typedef.type_declarators, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -586,7 +657,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		if (ret)
 			return ret;
 		nr_declarators = 0;
-		cds_list_for_each_entry(iter, &node->u.typealias_target.type_declarators, siblings) {
+		bt_list_for_each_entry(iter, &node->u.typealias_target.type_declarators, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -617,7 +688,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		if (ret)
 			return ret;
 		nr_declarators = 0;
-		cds_list_for_each_entry(iter, &node->u.typealias_alias.type_declarators, siblings) {
+		bt_list_for_each_entry(iter, &node->u.typealias_alias.type_declarators, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -657,6 +728,9 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		case NODE_STRING:
 		case NODE_ENUMERATOR:
 		case NODE_ENUM:
+		case NODE_CLOCK:
+		case NODE_CALLSITE:
+		case NODE_ENV:
 		default:
 			goto errinval;
 		}
@@ -703,7 +777,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		case NODE_UNARY_EXPRESSION:
 			goto errperm;
 		}
-		cds_list_for_each_entry(iter, &node->u.floating_point.expressions, siblings) {
+		bt_list_for_each_entry(iter, &node->u.floating_point.expressions, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -718,7 +792,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 
 		}
 
-		cds_list_for_each_entry(iter, &node->u.integer.expressions, siblings) {
+		bt_list_for_each_entry(iter, &node->u.integer.expressions, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -735,7 +809,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 			goto errperm;
 		}
 
-		cds_list_for_each_entry(iter, &node->u.string.expressions, siblings) {
+		bt_list_for_each_entry(iter, &node->u.string.expressions, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -756,7 +830,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		{
 			int count = 0;
 
-			cds_list_for_each_entry(iter, &node->u.enumerator.values,
+			bt_list_for_each_entry(iter, &node->u.enumerator.values,
 						siblings) {
 				switch (count++) {
 				case 0:	if (iter->type != NODE_UNARY_EXPRESSION
@@ -781,7 +855,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 			}
 		}
 
-		cds_list_for_each_entry(iter, &node->u.enumerator.values, siblings) {
+		bt_list_for_each_entry(iter, &node->u.enumerator.values, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -803,7 +877,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		if (ret)
 			return ret;
 
-		cds_list_for_each_entry(iter, &node->u._enum.enumerator_list, siblings) {
+		bt_list_for_each_entry(iter, &node->u._enum.enumerator_list, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -822,7 +896,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 			node->u.struct_or_variant_declaration.type_specifier_list);
 		if (ret)
 			return ret;
-		cds_list_for_each_entry(iter, &node->u.struct_or_variant_declaration.type_declarators, siblings) {
+		bt_list_for_each_entry(iter, &node->u.struct_or_variant_declaration.type_declarators, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -838,7 +912,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		case NODE_UNARY_EXPRESSION:
 			goto errperm;
 		}
-		cds_list_for_each_entry(iter, &node->u.variant.declaration_list, siblings) {
+		bt_list_for_each_entry(iter, &node->u.variant.declaration_list, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
@@ -855,7 +929,7 @@ int _ctf_visitor_semantic_check(FILE *fd, int depth, struct ctf_node *node)
 		case NODE_UNARY_EXPRESSION:
 			goto errperm;
 		}
-		cds_list_for_each_entry(iter, &node->u._struct.declaration_list, siblings) {
+		bt_list_for_each_entry(iter, &node->u._struct.declaration_list, siblings) {
 			ret = _ctf_visitor_semantic_check(fd, depth + 1, iter);
 			if (ret)
 				return ret;
